/*

JAVASCRIPT OBJECT BROWSER v3.0 RC (c) 2001-2004 Angus Turnbull. All Rights Reserved.
Visit http://www.twinhelix.com for licensing info and more scripts!
Updated: 2 July 2004


-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

Can index, browse & alter parts of the Document Object Model in a table,
in real time, displaying the names and values of a browser's page objects.
Use this to assist in development -- it's perfect for answering questions
like "How do I find the height of..." when you're writing DHTML. It's a
little rough in places, and may not be able to browse all parts of the DOM.
Your mileage may vary, especially in older/unusual browsers.
   Not all objects seem to be indexed properly, such as document.layers and
document.plugins in NS4, although you can often use the 'Go To' box at the
top of the screen to browse them, try typing "document.layers.0".
   You can also browse through any objects you have created in another page --
again, type in their names -- and functions like "document.getElementById()".
Oh, and find a recent copy of Mozilla to browse, it lists functions of objects
as well as their properties which is very useful indeed :).


Known Issues / Rare and Extraordinary Bugs:

* Under alternative shells (such as LiteStep) or alternative browsers using the
  IE rendering engine (tabbed browsers etc), IE sometimes throws more errors
  than normal when you try to browse some parts of the object hierarchy.
  - Workaround: 'System Integration', huh? Better boot to Explorer and run
    the plain vanilla IEXPLORE.EXE browser UI for best results.

* You can browse another file by opening this page from a window.open() call
  or opening another page from this one. However, once you close that page,
  you cannot browse further as the link between windows is broken.
  - Workaround: Refresh this page after closing the external file. You could
    modify the code to check for 'top.cwn.closed' but then you'd be unable to
    pass object references to the browse function anyway.


I hope you find this useful. Try loading in an IFRAME or similar, as this script
will automatically detect and browse the topmost frameset - this can help you
debug a page you're working on, or just locate those elusive .style properties :).
Also calling via window.open() works well (the script will automatically select
the parent window's topmost frameset to browse), and you can enter a filename
down the bottom which will be opened and browsed from the current script window.
The "window.location" object will of course tell you the file being browsed.


You can add this to the IE context menu by following these steps (Power users
only -- if you aren't familiar with the Registry, skip this):
 1) Open the Registry Editor (Start -> Run -> 'regedit.exe')
 2) Browse to HKEY_CURRENT_USER\Software\Microsoft\Internet Explorer\MenuExt
 3) Add a new key called 'Browse DOM' or something similar. Open it.
 4) Set its 'Default' value string to the path to this browser HTML file,
    for example 'c:\stuff\objbrowser.html'.
 5) Add another new 'Binary' value called 'contexts', set to '01'.
 6) Reload IE and try it out.

Otherwise, this script supports activation via bookmarklets, which is probably
more useful as you can browse any domain with the JS file on you hard disk.
Click the link at the bottom of the output for instructions.


Ideas would be welcome on better classification of event handlers. Most are
actually of type 'object' (in IE at least), but some have no type. Currently,
it's called an event handler if it starts with 'on' and is not 'onLine' from IE's
navigator objects. Any ideas as to how this can be improved?

Also, it doesn't work in Opera v5 or 6, as they don't support enumerating built in
objects according to my experiments and their technical docs. Opera 7+ should be OK.
   I had to restrict innerHTML et. al. in IE4 as on my system it generated errors.
That could always be because I have IE5 and IE6 installed as well though (yup, it's
possible, with some DLL wrangling).

I have managed to royally crash IE by browsing some sites with this, so probably
make sure you're not doing anything really important when testing for a start.
You use it entirely at your own risk, on an "as-is" basis with no warranties, etc.


Update History:

v3.0: * Converted main code over to JS file.
      * Added bookmarklet capability, so it can run on foreign domains.
      * Colourised property results by value type (string, number, boolean, other).
      * Wrapped code up in an object, to reduce global vars.
      * Moved bad object lists outside the browse function, and gave the script a
        BIG speed boost when testing properties as they're now associative arrays.
      * Now there's try/catch error checking for object values too.
      * Altered "set property" function to make it more sensible & set proper types.
      * Enabled sorting in all browsers, might as well...
      * Put in a small referencing bug workaround to make it go in Opera 7.
      * Worked around a weird Safari bug I still don't understand.
      * Misc. cleaning up and tweaks, polished HTML output a bit, e.g. now object
        titles are displayed in browser back/forward listing.

v2.1: * Made IE6 sort categories as well, something MS must have forgotten.
      * Built in try/catch error checking in new browsers, for less errors.
      * Fixed altering properties in arrays, e.g. document.all[0].something.
      * Added form to evaluate a JS string in the window being browsed.
      * A general trimming and tweaking.
      * v2.11 Maintenance Release: Fixed frameElement as property crash in IE6,
        also fixed window.fullScreen uncaught exception in Mozilla 1.x.

v2.0  * Passed object reference as well as strings to objBrowse().
      * Slimmed down bad object checking, removed 'restricted' mode.
      * Made it browse collections properly as obj[0] not obj.0
      * Made frame usage much simpler & more reliable.
      * Added a nice 'sorry' error dialog for IE at least.
      * Made NS6 sort categories into alphabetical order.
      * Displays the first 100 chars of an object's value only, makes browsing
        easier and cleaner when big functions are assigned to events.
      * Loads of other tweaks and changes. 

v1.0  * Initial public release.

*/








// Attempt to be courteous even when chronically/perpetually broken.
window.onerror = function(err, file, line) {
 alert('The Object Browser has encountered an internal error,\n' +
  'possibly due to to an unevaluable property.\n\n' +
  'Error in Line ' + line + ': ' + err + '.\n\n' +
  'Property name: ' + JSObj.propName + '\n\n' +
  'Please select a different object to browse, or refresh.');

 return true;
}



// For the rest of the script, I'll wrap up the functions in an object, to stop polluting
// the place with global vars and make something interesting to browse.
// Since there's only one of these, I can't be bothered prototyping it.

window.JSObj = new Object();

// Records the property the script is currently attempting to evaluate.
JSObj.propName = '';

// From where do we start browsing? preObj is a variable that is prepended to any text
// string passed to the object browser. Try and find an opener window, or just a frameset.
JSObj.preObj = (top.opener ? 'top.opener.' : 'top.');


// Some browsers can't detect properties of objects well and crash the script.
// So here are lists of values to skip or force to a particular type.

// window.fullScreen causes problems in some new Mozilla versions, so ignore it entirely.
JSObj.skipProps = { fullScreen:true }


// Some browsers can't reliably detect some types, so preset these property types.
JSObj.forceTypes = { height:'number', innerHTML:'string', innerText:'string',
 outerHTML:'string', outerText:'string', applets:'object', embeds:'object' };


// Depending on the browser, we might want to force some objects' values as they cannot
// be detected by the script normally. IE needs a hand on the following:
if (document.all) JSObj.forceValues = { applets:'[object]', clientInformation:'[object]',
 embeds:'[object]', external:'[object]', history:'[object]', mimeTypes:'[object]',
 navigator:'[object]', opsProfile:'[object]', plugins:'[object]', userProfile:'[object]' };

// NS4 has its own set of dodgy objects. Mozilla is pretty good, though.
if (document.layers) JSObj.forceValues = { applets:'[object]', embeds:'[object]' };






// Takes a string like document.all.0.style and formats to document.all[0].style,
// which makes it evaluable without JS errors.
JSObj.getSafeObj = function(str)
{
 var arr = str.split('.');
 str = '';
 for (var i = 0; i < arr.length; i++)
 {
  if (arr[i] != parseInt(arr[i])) str += arr[i] + '.';
  else str = str.substring(0, str.length - 1) + '[' + arr[i] + '].';
 }
 return str.substring(0, str.length - 1);
}

// Returns a string to be evaluated, wrapped in try/catch blocks in supported browsers.
JSObj.tryString = function(try_str, catch_str)
{
 if (document.getElementById) return 'try{' + try_str + '}catch(err){' + catch_str + '}';
 else return try_str;
}





// Main browsing function, pass it an object name and optionally its reference.
JSObj.browse = function(objText, objRef) { with (this)
{
//alert('JSObj.browse() called with args: ' + objText + ', ' + objRef);

 // *** IF NO REFERENCE, CLEAN UP OBJECTS AND EVALUATE ***
 // Trap leading/trailing spaces, periods etc...
 if (!objRef) while(1)
 {
  // Has this loop wound up with nothing left?
  if (!objText) objText = 'window';

  var badChar = false;

  // Check first char for bad character.
  if ((objText.substring(0, 1) == ' ') || (objText.substring(0, 1) == '.'))
  {
   objText = objText.substring(1);
   badChar = true;
   continue;
  }

  // Check last char for bad character.
  if ((objText.substring(objText.length - 1) == ' ') ||
      (objText.substring(objText.length - 1) == '.'))
  {
   objText = objText.substring(0, objText.length - 1);
   badChar = true;
   continue;
  }

  // First and last chars OK? Might as well go on, any other errors can be refreshed.
  if (!badChar)
  {
   // Why the first eval? I don't know either. Ask the NS4 dev team why it's needed...
   eval('window');
   objRef = eval(getSafeObj(preObj + objText));
   break;
  }
 }


  // *** LOOP & DETECT BAD PROPERTIES, CATEGORISE PROPERTIES INTO ARRAYS ***


 // Multidimensional array to hold 3 classes of names of children/properties etc.
 var Objects = new Array(new Array(), new Array(), new Array());

 // Names of the categories.
 var catNames = new Array('Child Objects','Properties','Event Handlers');

 // We have to actually access some objects to make NS4 show them :).
 if (navigator.appName=='Netscape' && !document.getElementById)
 {
  if (eval(preObj+'screen')) {}
  if (eval(preObj+'navigator')) {}
  if (eval(preObj+'document.layers')) {}
 }

 // The main loop.
 for (propName in objRef)
 {
  //if (objText.indexOf('document') != -1) alert('Categorising: ' + propName);

  var objType = '', aIndex, aRef;

  // Skip displaying some objects.
  if (JSObj.skipProps && JSObj.skipProps[propName]) continue;

  // Some properties cause crashes, so preset them...
  if (JSObj.forceTypes && JSObj.forceTypes[propName]) objType = forceTypes[propName];

  // Well, if it hasn't been ruled out, call typeof()...
  // If the script crashes here, add the property to the JSObj.forceTypes array.
  if (!objType) eval(tryString('objType = typeof objRef[propName]', 'objType = "string"'));


  // Start categorising into arrays... aIndex is the type of propName.

  // Call it an event handler as their objTypes vary, add to end of Objects[2]...
  // I know this is a bit dodgy, anyone know a better way of detecting handlers?
  if (propName.substring && (propName.substring(0, 2) == 'on') && (propName != 'onLine'))
   aIndex = 2;

  // Otherwise add to 'objects' subarray if it reports itself as an object.
  else if (objType == 'object') aIndex = 0;

  // Otherwise call it a property and move on.
  else aIndex = 1;

  // Create a new object within the Objects[] array, to which we'll add data.
  var aRef = Objects[aIndex][Objects[aIndex].length] = new Object();
  aRef.type = objType;
  aRef.name = propName;


  // Now the hard bit... detect the value of the object, and set as aRef.value.

  // Check if this property has a forced value, if it does set it and loop again.
  if (JSObj.forceValues && JSObj.forceValues[propName]) aRef.value = forceValues[propName];
  else
  {
   var str = '';
   // Windows NT4 seems to crash on 'domain' for some reason under IE5.
   // IE6 dies horribly when window.frameElement is a property too.
   if ((propName != 'domain') && ((propName != 'frameElement') || (objType == 'object')))
   {
    var errStr = '????';

    // IE4 seems to throw fits when being made to access innerHTML etc. So skip 'em.
    if ((document.all && !document.getElementById) &&
        (propName=='innerHTML' || propName=='innerText' || propName=='outerHTML' ||
         propName=='outerText'))
     str = errStr;

    // OK, if the script is crashing about here, add this property to JSObj.forceValues
    eval(tryString('str = new String(objRef[propName])', 'str = errStr'));
   }
   
   // Lose angle brackets, otherwise innerHTML etc. look really weird :).
   str = str.replace(/</g, '&lt;').replace(/>/g, '&gt;');

   // Trim it down and record the final value in the array.
   if (str.length > 100) str = str.substring(0, 100) +
    ' <span style="font-weight: bold; color: #009900">...</span>';
   aRef.value = str;
  }

  // Move onto next object in for() loop...
 }


 // All hands on deck for a dirty old Mozilla hack. Since some versions (1.2-1.3ish)
 // won't list window.document, we really have to add it manually as it's pretty useful.
 if (navigator.userAgent.indexOf('Gecko')>-1 && objRef && objRef.navigator)
 {
  var docFound = false;
  for (var count = 0; count < Objects[0].length; count++)
   if (Objects[0][count].name == 'document') { docFound = true; break }
  if (!docFound) Objects[0][Objects[0].length] =
   { name:'document', type:'object', value:'[object HTMLDocument]' }
 }

 // Likewise, help out NS4 with some document.layers fakery.
 if (objRef && objRef.layers)
 {
  var lyrFound = false;
  for (var count = 0; count < Objects[0].length; count++)
   if (Objects[0][count].name == 'layers') { lyrFound = true; break }
  if (!lyrFound) Objects[0][Objects[0].length] =
   { name:'layers', type:'object', value:'[object LayerArray]' }
 }






   // *** WRITE HEADING OF OUTPUT WINDOW ***

 // Find our output document. Opera 7.01 and 7.02 are a bit stupid and throw security errors
 // when I attempt to write to "outF.document", so here's a workaround using the neat old
 // DOM getElementsByName function (ironically, I used this script to get the right one :).
 var doc = window.opera ? document.getElementsByName('outF')[0].contentDocument : outF.document;

 // Clear anything already written to output document.
 doc.close();
 doc.open('text/html');

 // The title of this frameset document, not the output frame.
 document.title = objText + ' - Object Model Browser';

 doc.writeln('<html>\n<head>\n<title>'+document.title+'</title>\n' +
  '<style type="text/css"><!--\n' +
  ' td { font: 8pt/10pt Verdana, Arial }\n' +
  ' .prop { color: #000099; text-decoration: none }\n' +
  '--></style>\n</head>\n');

 // Start the body, and prepare to write the top links...
 // Yeah, I know the marginxyz attributes aren't kosher, but NS4 doesn't like CSS much.
 doc.writeln('<body onLoad="document.forms.goTo.text.select()" ' +
  'marginwidth="10" marginheight="10" leftmargin="10" topmargin="10">');
 doc.writeln('<div align="center">\n');

 doc.writeln('<h2>\n Object:');

 // Split up object name into array of pieces...
 var objSplit = objText.split('.');
 var wholeName = '';

 for (var count = 0; count < objSplit.length; count++)
 {
  // ...And recombine bit by bit.
  wholeName += objSplit[count];
  doc.write('<a href="javascript:parent.JSObj.browse(\'' + wholeName + '\', ' +
   getSafeObj(preObj + wholeName) + ')">' + objSplit[count] + '</a>');
  if (objSplit[count + 1])
  {
   doc.write('.');
   wholeName += '.';
  }
 }
 doc.writeln('\n</h2>\n');

 // Don't want form resubmitting, so call objBrowse function & return false.
 // Pass only the text, the reference will be sorted out by this function.
 doc.writeln('<form name="goTo" onSubmit="setTimeout(\'parent.JSObj.browse(' +
  'document.forms.goTo.text.value)\', 10); return false">\n' +
  ' <b>Go To:</b> <input name="text" type="text" size="50" value="' + objText + '">\n' +
  ' <input type="submit" value="Go / Refresh">\n</form>\n');


   // *** DISPLAY THE OBJECT TYPES IN A TABLE ***


 // Define colours used for properties.
 var propCols = new Object();

 // Define the colours used for various object types.
 // NS4 doesn't like setting propCols.boolean as it's kind of a reserved word :).
 propCols['string'] = '#990000';
 propCols['number'] = '#006600';
 propCols['boolean'] = '#000099';


//alert('Writing the properties...');
 doc.writeln('<table border="1" cellpadding="1" cellspacing="1">');

 for (var objType = 0; objType < 3; objType++)
 {
  // Write grey table heading using catNames[]
  doc.writeln(' <tr><td colspan="2" bgcolor="#CCCCCC" align="center"><b>' + catNames[objType] +
   '</b></td></tr>');

  // Sort them into alphabetical order then loop through.
  // Don't apply sort to collections (.length property), but do apply to windows.
  if (objRef && (!objRef.length||objRef.navigator))
   Objects[objType].sort(new Function('a', 'b',
    'return (a.name.toLowerCase() > b.name.toLowerCase() ? 1 : -1)'));


  // No objects in this category?
  if (!Objects[objType].length)
  {
   doc.writeln(' <tr><td colspan="2" align="center">There are no indexed ' + catNames[objType] +
    ' for this Object.</td></tr>');
  }
  else for (var propCount = 0; propCount < Objects[objType].length; propCount++)
  {
   // Reference to this array entry, and set propName for debugging help.
   var propName = Objects[objType][propCount].name;

   // Write out a table row, first cell containing object's proper name (skip leading chars).
   wholeName = objText + '.' + Objects[objType][propCount].name;
   doc.write(' <tr><td>');

   if (objType == 0)
   {
    // It's an object, so write a link tag, we can browse into this.
    doc.write('<a href="javascript:parent.JSObj.browse(\'' + wholeName +
     '\', ' + getSafeObj(preObj + wholeName) + ')">' + wholeName + '</a>');
   }
   else
   {
    // Just write the name, it's a property or event handler.
    doc.write('<a class="prop" href="javascript:parent.JSObj.alterProp(\'' + wholeName +
     '\')">' + wholeName + '</a>');
   }

   // Write its value from the array -- use a space for empty objects.
   with (Objects[objType][propCount])
    doc.writeln('</td><td ' +
     ((type && objType==1) ? 'style="color: '+propCols[type]+'"' : '') + '>' +
     ((value != '') ? value : '&nbsp;') + '</td></tr>')
  }
 }


 // Page footer.
 doc.writeln('</table>\n\n<form name="openFile" onSubmit="setTimeout(\'' +
  'parent.JSObj.loadURL(document.forms.openFile.fileName.value)\', 10); return false">\n' + 
  ' Open: <input name="fileName" type="file"></input>\n' +
  ' <input type="submit" value="Open Selection..."></input>\n</form>\n');


 // Let's give us the ability to type in some JavaScript.
 doc.writeln('<form onsubmit="with (this.txt){'+preObj+'eval(value);focus();select();' +
  'return false}">\n JavaScript eval(): <input name="txt" type="text" size="50"></input>\n</form>\n');

 // The bookmarklet link, use double quotes for obvious reasons here.
 var bkmk = "javascript:var JSObjPath='file:///C:/Windows/JSObj.js';"+
  "with(window.open('','','menubar=1,toolbar=1,location=0,status=1,scrollbars=1,resizable=1').document)"+
  "{open('text/html');write('<html><script type=\\\'text/javascript\\\' src=\\\''+JSObjPath+'\\\'></script>"+
  "<noframes>Error, JSOBJ.JS cannot be located!</noframes></html>');close()}";

 doc.writeln('<p><a href="'+bkmk+'" onclick="alert(\'To use: Drag onto links toolbar or bookmarks.\\n' +
  'Click OK to the warning, then edit the link to point at the correct JSOBJ.JS file.\\n' +
  'You can then click that link to browse the DOM of any page on any domain.\'); return false">' +
  'JSObj</a> (bookmarklet; click for instructions).</p>\n');

 doc.writeln('\n</div>\n</body>\n</html>');
 // That's it, tell browser it's finally 'loaded'.
 doc.close();
}}


   // *** OPEN A CHILD WINDOW TO BROWSE ***

JSObj.loadURL = function(fName) { with (this)
{
 // Fix up the URL for Netscape 4.
 if ((fName.substring(0,4) != 'http') && (fName.substring(0,4) != 'file'))
  fName = 'file:///' +fName;

 // Open a window there, set it as the root for browsing, and browse it.
 top.JSObjWin = top.open(fName,'',
  'menubar=yes,toolbar=yes,location=yes,directories=no,status=yes,resizable=yes,scrollbars=yes');
 preObj = 'top.JSObjWin.';
 browse('window');
}}


   // *** PROPERTY CLICK HANDLER ***

// Gets the new value for a property, sets it, then refreshes.
JSObj.alterProp = function(objText) { with (this)
{
 var newVal = prompt('Enter new value for: ' + objText, eval(getSafeObj(preObj + objText)));
 if (typeof newVal == 'string')
 {
  // Parse the full text into object/property and get reference.
  var objName = objText.substring(0, objText.lastIndexOf('.'));
  var propName = objText.substring(objText.lastIndexOf('.') + 1);
  var objRef = eval(getSafeObj(preObj + objName));

  // Now decide what type we're setting -- use string only as the default.
  if (newVal == parseFloat(newVal).toString()) objRef[propName] = parseFloat(newVal);
  else if (newVal=='true') objRef[propName] = true;
  else if (newVal=='false') objRef[propName] = false;
  else if (newVal=='null') objRef[propName] = null;
  else objRef[propName] = newVal;
  browse(objName);
 }
}}





// *** MAIN SCRIPT EXECUTION STARTS HERE ***




// *** IE CONTEXT MENU COMPATIBILITY ***

if (document.all && window.external && external.menuArguments && external.menuArguments!=self)
{
 // Instructs the caller to open a new window to this location so top.opener
 // is detected next time, then exit as this connection is not persistent.
 external.menuArguments.open(window.location.href, '',
  'menubar=1,toolbar=1,location=0,status=1,scrollbars=1,resizable=1');
}


// *** OTHERWISE START OBJECT BROWSER SCRIPT WITH FRAMESET ***

else
{
 // Create a frameset within the current document, which will run the browse function onload.
 document.write('<frameset rows="100%,*" resizable="no" frameborder="0" border="0" ' +
  'onload="JSObj.browse(\'window\')"><frame name="outF" src="about:blank"></frame>\n</frameset>');
}

